if not exist tests\data mkdir tests\data || exit 1
xcopy test_data\coutwildrnp.zip tests\data /s /e /y || exit 1
xcopy test_data\coutwildrnp.tar tests\data /s /e /y || exit 1
xcopy test_data\coutwildrnp.json tests\data /s /e /y || exit 1

:: The tests are finishing but it hangs at the end :-/
REM pytest -s -rxs -v -k "not (test_fio_ls_single_layer or test_directory or test_directory_trailing_slash or test_options or test_transaction)" tests



fio --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
fio ls test_data/test.shp
IF %ERRORLEVEL% NEQ 0 exit /B 1
fio info test_data/test.shp
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
